/* **********************************************************************
    Copyright 2003 Rensselaer Polytechnic Institute.

    All worldwide rights reserved. A license to use, copy, modify and
    distribute this software for noncommercial research purposes only is
    hereby granted, provided that this copyright notice and accompanying
    disclaimer is not modified or removed from the software.

    DISCLAIMER: The software is distributed" AS IS" without any express or
    implied warranty, including but not limited to, any implied warranties
    of merchantability or fitness for a particular purpose or any warrant)'
    of non-infringement of any current or pending patent rights. The authors
    of the software make no representations about the suitability of this
    software for any particular purpose. The entire risk as to the quality
    and performance of the software is with the user. Should the software
    prove defective, the user assumes the cost of all necessary servicing,
    repair or correction. In particular, neither Rensselaer Polytechnic
    Institute, nor the authors of the software are liable for any indirect,
    special, consequential, or incidental damages related to the software,
    to the maximum extent the law permits.
*/

package edu.rpi.sss.util;

import edu.rpi.sss.util.fmt.FmtUtil;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Properties;

/**
 *
 * @author  Mike Douglass
 * @version
 * PropertyInfo.java
 *
 * Created on August 4, 1999, 10:33 AM
 *
 * Class and methods to help maintain a set of program properties.
 */

public class PropertyUtil extends Properties {

  // HashTables (and Properties) cannot hold null values so we keep a table
  //               of all names we ever knew.
  private Hashtable allNames;
  private Integer dummyVal = new Integer(0);

  public PropertyUtil() {
    allNames = new Hashtable();
  }

  public PropertyUtil(PropertyUtil defaults) {
    super(defaults);
    allNames = new Hashtable();
  }

  /** Create a set of properties with defined defaults.
   */
  public PropertyUtil(String[] names, String[] defaults) throws PropertyUtilException {
    this();
    put(names, defaults);
  }

  /** Create a set of properties with defined defaults.
   */
  public PropertyUtil(String[][] defaults) throws PropertyUtilException {
    this();
    put(defaults);
  }

  public PropertyUtil(Hashtable tbl) throws PropertyUtilException {
    this();
    put(tbl);
  }

  public void put(String name, String val) {
    if (val == null) {
      remove(name);
    } else {
      setProperty(name, val);
    }
    allNames.put(name, dummyVal);
  }

  public void put(String[][] namevals) throws PropertyUtilException {
    for (int i = 0; i < namevals.length; i++) {
      if (namevals[i][0] != null) {
        put(namevals[i][0], namevals[i][1]);
      }
      allNames.put(namevals[i][0], dummyVal);
    }
  }

  public void put(String[] names, String[] vals) throws PropertyUtilException {
    if (names.length != vals.length) {
      throw new PropertyUtilException("vals size doesn't match property list size");
    }

    for (int i = 0; i < names.length; i++) {
      if (vals[i] != null) {
        put(names[i], vals[i]);
      }
      allNames.put(names[i], dummyVal);
    }
  }

  /** Transfer all the name/value pairs from the hashtable.
   */
  public void put(Hashtable tbl) {
    Enumeration e = tbl.keys();
    while (e.hasMoreElements()) {
      String name = (String)e.nextElement();
      String val = (String)tbl.get(name);

      put(name, val);
    }
  }

  public synchronized Object get(Object key) {
//  public String get(String name) {
    return getProperty((String)key);
  }

  public String getStr(String key) {
    return getProperty(key);
  }

  public String get(String name, String defaultVal) {
    return getProperty(name, defaultVal);
  }

  /** If the named property is present and has a value use that.
      Otherwise, set the value to the given default and use that.
   */
  public String checkProp(String name, String defaultVal) {
//    String val = get(name);
    String val = (String)get(name);

    if (val == null) {
      put(name, defaultVal);
      val = defaultVal;
    }

    return val;
  }

  /** Ensure we have a default value for named properties.
   */
  public void setDefault(String[] names, String[] defaultVals)
      throws PropertyUtilException {
    if (names.length != defaultVals.length) {
      throw new PropertyUtilException("vals size doesn't match property list size");
    }

    for (int i = 0; i < names.length; i++) {
      if (get(names[i]) == null) {
        put(names[i], defaultVals[i]);
      }
    }
  }

  /** Ensure we have a default value for named properties.
      Doesn't set the value if it is already set.
   */
  public void setDefault(String[][] nameVals) throws PropertyUtilException {
    for (int i = 0; i < nameVals.length; i++) {
      String[] nameVal = nameVals[i];
      if (get(nameVal[0]) == null) {
        put(nameVal[0], nameVal[1]);
      }
    }
  }

  public int getInt(String name) throws PropertyUtilException {
    try {
      return Integer.parseInt(getProperty(name), 10);
    } catch (NumberFormatException e) {
      throw new PropertyUtilException("Invalid integer value: " + e.getMessage());
    }
  }

  public int getInt(String name, int defaultVal) throws PropertyUtilException {
    String val = getProperty(name);

    if (val == null) {
      return defaultVal;
    }

    try {
      return Integer.parseInt(val, 10);
    } catch (NumberFormatException e) {
      throw new PropertyUtilException("Invalid integer value: " + e.getMessage());
    }
  }

  public boolean getBool(String name) throws PropertyUtilException {
    return getBool(name, null);
  }

  public boolean getBool(String name, boolean defaultVal) throws PropertyUtilException {
    if (defaultVal) {
      return getBool(name, "true");
    }

    return getBool(name, "false");
  }

  private boolean getBool(String name, String defaultVal) throws PropertyUtilException {
    String val = getProperty(name, defaultVal);

    if (val == null) {
      val = defaultVal;
    }

    if (val == null) {
      throw new PropertyUtilException("Invalid boolean value: null");
    }

    val = val.toLowerCase();
    if ((val.equals("true")) || (val.equals("yes"))) {
      return true;
    }

    if ((val.equals("false")) || (val.equals("no"))) {
      return false;
    }

    throw new PropertyUtilException("Invalid boolean value: not one of true, false, yes or no");
  }

  public void setBool(String name, boolean val) throws PropertyUtilException {
    if (val) {
      put(name, "true");
    } else {
      put(name, "false");
    }
  }

  /* The Configurator class is passed to the configure method to allow
     it to retrieve property values from the environment in which the
     PropertyUtil class is being used. The get method will be called for each
     property which we know about. A null return means don't change the
     property value.
   */
  public abstract static class Configurator {
    public abstract String get(String name);
  }

  public void configure(Configurator c) {
    Enumeration e = allNames.keys();

    while (e.hasMoreElements()) {
      String name = (String)e.nextElement();
      String val = c.get(name);

      if (val != null) {
        put(name, val);
      }
    }
  }

  public Object clone() {
    PropertyUtil np = new PropertyUtil();

    Enumeration e = propertyNames();
    while (e.hasMoreElements()) {
      String name = (String)e.nextElement();
      String val = getProperty(name);
      np.put(name, val);
    }

    return np;
  }

  /** Replace ${<name>} constructs with the named property value
      Note this should allow us to escape the ${ sequence and we may run into
      conflicts with format strings (which use {...})
      I should really go through the string char by char and check for matching
      brackets.

      We also look for $${ and $}. These are replaced with ${ and }.
      These allow a pass through the conversion process such as happens when
      properties are read from an ldap directory and processed with this routine
      before storing in this object.

      Note, this routine will not accept nested constructs nor will it handle
      mismatched bracket too well.
   */

  // States we can be in
  public static final int achar = 0;   // We just saw some random char
  public static final int dollar = 1;  // We saw a dollar
  public static final int dollar2 = 2; // We saw 2 dollars
  public static final int needrbracket = 3; // We're waiting for }

  // Action index
  public static final int actNull = 0;
  public static final int actAppendCurChar = 1;
  public static final int actStartNeedRbracket = 2;
  public static final int actAppendStartPar = 3;
  public static final int actAppendDollarParEnd = 4;
  public static final int actAppendParValue = 5;
  public static final int actAppendDollarCurChar = 6;

  // char types
  public static final int chany = 0;
  public static final int chdollar = 1;
  public static final int chlbracket = 2;
  public static final int chrbracket = 3;

  // Indexed by charType, curState [0] for next state
  // Indexed by charType, curState [1] for action
  public static final int[][][] stateTab = {
    // chany
    {
      {achar, actAppendCurChar},
      {achar, actAppendDollarCurChar},
      {achar, actAppendCurChar},
      {needrbracket, actNull},
    },

    // chdollar
    {
      {dollar, actNull},
      {dollar2, actNull},
      {dollar2, actAppendCurChar},
      {needrbracket, actNull},
    },

    // chlbracket
    {
      {achar, actAppendCurChar},
      {needrbracket, actStartNeedRbracket},
      {achar, actAppendStartPar},
      {needrbracket, actNull},
    },

    // chrbracket
    {
      {achar, actAppendCurChar},
      {achar, actAppendCurChar},
      {achar, actAppendDollarParEnd},
      {achar, actAppendParValue},
    },
  };

  public String replacePars(String val) {
    if (val == null) {
      return null;
    }

    int state = achar;
    StringBuffer sb = new StringBuffer();
    int nameStart = 0;

    for (int i = 0; i < val.length(); i++) {
      char c = val.charAt(i);
      int ctype;
      if (c == '$') {
        ctype = chdollar;
      } else if (c == '{') {
        ctype = chlbracket;
      } else if (c == '}') {
        ctype = chrbracket;
      } else {
        ctype = chany;
      }

      int nextState = stateTab[ctype][state][0];

      /*
        System.out.println("ch=" + val.substring(i, i+1) +
          " state = " + state + " nextState = " + nextState + "\n" +
          "     action = " + stateTab[ctype][state][1]);
          */

      switch (stateTab[ctype][state][1]) {
        case actNull: {
          // Do nothing
          break;
        }
        case actAppendCurChar: {
          sb.append(c);
          break;
        }
        case actAppendDollarCurChar: {
          sb.append('$');
          sb.append(c);
          break;
        }
        case actStartNeedRbracket: {
          nameStart = i + 1;
          break;
        }
        case actAppendStartPar: {
          sb.append("${");
          break;
        }
        case actAppendDollarParEnd: {
          // We saw $$}
          sb.append("$}");
          break;
        }
        case actAppendParValue: {
          // System.out.println("get prop " + val.substring(nameStart, i));
          sb.append(getStr(val.substring(nameStart, i)));
          break;
        }
      }

      state = nextState;
    }

    if (state == dollar) {
      // trailing dollar
      sb.append('$');
    } else if (state == dollar2) {
      // 2 trailing dollars - assume escaped
      sb.append('$');
    }

    return sb.toString();
  }

  /** Some useful formatting routines
   */

  // Format a message consisting of a format string saved as a property
  //  plus one string parameter
  public String fmtProp(String propName, String arg) {
    return FmtUtil.fmtMsg(getStr(propName), arg);
  }

  // Format a message consisting of a format string saved as a property
  //  plus two string parameters
  public String fmtProp(String propName, String arg1, String arg2) {
    return FmtUtil.fmtMsg(getStr(propName), arg1, arg2);
  }

  // Format a message consisting of a format string saved as a property
  //  plus 3 string parameters
  public String fmtProp(String propName, String arg1, String arg2, String arg3) {
    return FmtUtil.fmtMsg(getStr(propName), arg1, arg2, arg3);
  }

  // Format a message consisting of a format string saved as a property
  // plus one integer parameter
  public String fmtProp(String propName, int arg) {
    return FmtUtil.fmtMsg(getStr(propName), arg);
  }

}
