package edu.washington.cac.calfacade.shared;

import java.io.Serializable;
import java.util.Set;

/** Value object representing the sponsor of an event
 *
 * @author Mike Douglass
 * @version 1.0
 */
public class SponsorVO implements Comparable,Serializable {
  private int id;
  private String name, phone, email, link;

  /** The person who created the sponsor */
  private UserVO creator;

  private boolean isPublic;

  /** Events referencing this sponsor
   */
  private Set events;

  public SponsorVO() {
  }

  /** Create a sponsor specifying all fields
   *
   *  @param id            int unique ID for the sponsor
   *  @param name           String Name of the sponsor
   *  @param phone          String Phone number
   *  @param email          String E-mail
   *  @param link           String URL with more info
   *  @param creator        UserVO person who created the event
   *  @param isPublic       boolean true if the event is viewable by everyone
   */
  public SponsorVO(int id,
                String name,
                String phone,
                String email,
                String link,
                UserVO creator,
                boolean isPublic) {
    this.id = id;
    this.name = name;
    this.phone = phone;
    this.email = email;
    this.link = link;
    this.creator = creator;
    this.isPublic = isPublic;
  }

  public void setId(int val) {
    id = val;
  }

  public int getId() {
    return id;
  }

  /** Set the name
   *
   * @param val    String name
   */
  public void setName(String val) {
    name = val;
  }

  /** Get the name
   *
   * @return String   name
   */
  public String getName() {
    return name;
  }

  public void setPhone(String val) {
    phone = val;
  }

  public String getPhone() {
    return phone;
  }

  public void setEmail(String val) {
    email = val;
  }

  public String getEmail() {
    return email;
  }

  /** Set the sponsor's URL
   *
   * @param link   String new URL
   */
  public void setLink(String link) {
    this.link = link;
  }

  public String getLink() {
    return link;
  }

  public void setIsPublic(boolean val) {
    isPublic = val;
  }

  public boolean isPublic() {
    return isPublic;
  }

  public void setPublic(boolean val) {
    isPublic = val;
  }

  public boolean getPublic() {
    return isPublic;
  }

  /** Set the creator
   *
   * @param val     UserVO creator of the sponsor
   */
  public void setCreator(UserVO val) {
    creator = val;
  }

  /**
   *
   * @return UserVO    creator of the sponsor
   */
  public UserVO getCreator() {
    return creator;
  }

  /** Set the events - must be a set of EventVO
   *
   * @param val      Set of EventVO
   */
  public void setEvents(Set val) {
    events = val;
  }

  /** Get the events
   *
   * @param Set      of EventVO
   */
  public Set getEvents() {
    return events;
  }

  /* ====================================================================
   *                   Object methods
   * ==================================================================== */

  /**  Compares this object with the specified object for order.
   * Returns a negative integer, zero, or a positive integer as this object
   * is less than, equal to, or greater than the specified object.
   *
   */
  public int compareTo(Object o) {
    if (!(o instanceof SponsorVO)) {
      return -1;
    }

    if (equals(o)) {
      return 0;
    }

    SponsorVO that = (SponsorVO)o;

    int res = CalFacadeUtil.cmpObjval(getName(), that.getName());

    if (res == 0) {
      res = CalFacadeUtil.cmpObjval(getPhone(), that.getPhone());
    }

    if (res == 0) {
      res = CalFacadeUtil.cmpObjval(getEmail(), that.getEmail());
    }

    if (res == 0) {
      res = CalFacadeUtil.cmpObjval(getLink(), that.getLink());
    }

    if (res == 0) {
      res = CalFacadeUtil.cmpObjval(getCreator(), that.getCreator());
    }


    return CalFacadeUtil.cmpBoolval(getPublic(), that.getPublic());
  }

  /** The id does not take part in these comparisons because we're presumably
   * doing this to see if two objects are the same without doing a db retrieval.
   * That is, one of the objects in a comparison may not have the id set.
   *
   * <p>This is particularly important with caching persistance engines such
   * as hibernate.
   *
   * <p>This is relevant for hashcode as the equals method requires that
   * hashcode return equal values for both objects in a comparison that are
   * to be considered equal.
   *
   * <p>As we have both objects to hand in the equals method we can check to
   * see if both ids are non-zero and use them in the comparison.
   */
  public int hashCode() {
    int hc = 1;

    if (getName() != null) {
      hc *= getName().hashCode();
    }

    if (getPhone() != null) {
      hc *= getPhone().hashCode();
    }

    if (getEmail() != null) {
      hc *= getEmail().hashCode();
    }

    if (getLink() != null) {
      hc *= getLink().hashCode();
    }

    if (getCreator() != null) {
      hc *= getCreator().hashCode();
    }

    if (getPublic()) {
      hc *= 2;
    }

    return hc;
  }

  public boolean equals(Object obj) {
    if (this == obj) {
      return true;
    }

    if (!(obj instanceof SponsorVO)) {
      return false;
    }

    SponsorVO that = (SponsorVO)obj;

    if (!CalFacadeUtil.eqObjval(getName(), that.getName())) {
      return false;
    }

    if (!CalFacadeUtil.eqObjval(getPhone(), that.getPhone())) {
      return false;
    }

    if (!CalFacadeUtil.eqObjval(getEmail(), that.getEmail())) {
      return false;
    }

    if (!CalFacadeUtil.eqObjval(getLink(), that.getLink())) {
      return false;
    }

    if (!CalFacadeUtil.eqObjval(getCreator(), that.getCreator())) {
      return false;
    }

    return getPublic() == that.getPublic();
  }

  public Object clone() {
    if (id <= CalFacadeDefs.maxReservedSponsorId) {
      // It's a reserved immutable sponsor. Just return it.
      return this;
    }

    return new SponsorVO(id, name, phone, email, link,
                         (UserVO)creator.clone(), isPublic);
  }
}
