/* **********************************************************************
    Copyright 2005 Rensselaer Polytechnic Institute.

    All worldwide rights reserved. A license to use, copy, modify and
    distribute this software for noncommercial research purposes only is
    hereby granted, provided that this copyright notice and accompanying
    disclaimer is not modified or removed from the software.

    DISCLAIMER: The software is distributed" AS IS" without any express or
    implied warranty, including but not limited to, any implied warranties
    of merchantability or fitness for a particular purpose or any warrant)'
    of non-infringement of any current or pending patent rights. The authors
    of the software make no representations about the suitability of this
    software for any particular purpose. The entire risk as to the quality
    and performance of the software is with the user. Should the software
    prove defective, the user assumes the cost of all necessary servicing,
    repair or correction. In particular, neither Rensselaer Polytechnic
    Institute, nor the authors of the software are liable for any indirect,
    special, consequential, or incidental damages related to the software,
    to the maximum extent the law permits.
*/

package edu.rpi.cct.uwcal.caldav;

import edu.rpi.cct.uwcal.calsvci.CalSvcI;
import edu.rpi.cct.uwcal.icalendar.ComponentWrapper;
import edu.rpi.cct.uwcal.icalendar.IcalTranslator;
import edu.rpi.cct.webdav.servlet.shared.WebdavIntfException;
import edu.rpi.cct.webdav.servlet.shared.WebdavNsNode;
import edu.rpi.cct.webdav.servlet.shared.WebdavProperty;
import edu.rpi.sss.util.xml.QName;

import edu.washington.cac.calfacade.shared.EventVO;
import edu.washington.cac.calfacade.shared.GuidVO;

import net.fortuna.ical4j.model.Calendar;
import net.fortuna.ical4j.model.Component;
import net.fortuna.ical4j.model.component.VEvent;

import java.io.StringReader;
import java.io.Reader;
import java.util.Collection;
import java.util.Vector;

/** Class to represent an entity such as events in caldav.
 *
 *   @author Mike Douglass   douglm@rpi.edu
 */
public class CaldavComponentNode extends CaldavUWCalNode {
  private EventVO event;

  /** The Component object
   */
  private VEvent vevent;
  private Calendar ical;

  private String veventString;

  private ComponentWrapper comp;

  public CaldavComponentNode(CaldavURI cdURI, CalSvcI svci,
                             IcalTranslator trans, boolean debug) {
    super(cdURI, svci, trans, debug);

    collection = false;
    allowsGet = true;

    contentLang = "en";
    contentLen = -1;
    contentType = "text/calendar";
  }

  public Component getVevent() {
    return vevent;
  }

  /* ====================================================================
   *                   Property methods
   * ==================================================================== */

  /** This method is called before each setter/getter takes any action.
   * It allows the concrete implementation to defer some expensive
   * operation to just before the first call.
   *
   * @param content     boolean flag indicating if this is a content related
   *                    property - that is a property which requires fetching
   *                    and/or rendering the content
   */
  public void init(boolean content) throws WebdavIntfException {
    try {
      if ((event == null) && exists) {
        String idVal = cdURI.getId();

        if (idVal == null) {
          return;
        }

        /* If the id is a valid internal guid it's one of ours. */
        GuidVO guid = GuidVO.valueOf(idVal);

        int evid;

        if (svci.isInternalGuid(guid)) {
          evid = guid.getId();
        } else {
          // Look up the id
          evid = svci.lookupEventName(idVal);
        }

        if (debug) {
          debugMsg("Try to get event " + evid + " for name " + idVal);
        }

        if (evid <= 0) {
          event = null;
          exists = false;
        } else {
        // event = svci.getEvent(guid.getId());
          event = svci.getEvent(evid);
          if (event == null) {
            exists = false;
          } else {
            vevent = trans.toIcalEvent(event);
            ical = trans.toIcal(event);
            comp = new ComponentWrapper(vevent);

            creDate = comp.getCreated();
            lastmodDate = comp.getLastModified();
            name = comp.getSummary();
          }
        }

        if (content && (event != null) && (veventString == null)) {
          veventString = ical.toString();
          contentLen = veventString.length();
        }
      }
    } catch (Throwable t) {
      throw new WebdavIntfException(t);
    }
  }

  /** Return a collection of property objects
   *
   * @param ns      String interface namespace.
   * @return Collection (possibly empty) of WebdavProperty objects
   */
  public Collection getProperties(String ns) throws WebdavIntfException {
    Vector v = new Vector();
    String icalNS = CaldavUWDef.icalNamespace;

    addProp(v, ICalTags.summary, name);
    addProp(v, ICalTags.dtstart, comp.getDtstart());
    addProp(v, ICalTags.dtend, comp.getDtend());
    addProp(v, ICalTags.duration, comp.getDuration());
    addProp(v, ICalTags.transp, comp.getTransp());
    addProp(v, ICalTags.due, comp.getDue());
//    addProp(v, ICalTags.completed,        | date-time from RFC2518
    addProp(v, ICalTags.status, comp.getStatus());
//    addProp(v, ICalTags.priority,         | integer
//    addProp(v, ICalTags.percentComplete, | integer
    addProp(v, ICalTags.uid, comp.getUid());
    addProp(v, ICalTags.sequence, comp.getSequence());
//    addProp(v, ICalTags.recurrenceId,    | date-time from RFC2518
//    addProp(v, ICalTags.trigger,          | see below TODO

// FIX FIX FIX
    addProp(v, ICalTags.hasRecurrence, "0");
    addProp(v, ICalTags.hasAlarm, "0");
    addProp(v, ICalTags.hasAttachment, "0");

    return v;
  }

  /** Returns an InputStream for the content.
   *
   * @return Reader       A reader for the content.
   */
  public Reader getContent() throws WebdavIntfException {
    init(true);

    if (veventString == null) {
      return null;
    }

    return new StringReader(veventString);
  }

  /* ====================================================================
   *                   Private methods
   * ==================================================================== */

  private void addProp(Vector v, QName tag, Object val) {
    if (val != null) {
      v.addElement(new WebdavProperty(tag, String.valueOf(val)));
    }
  }
}
