/* **********************************************************************
    Copyright 2005 Rensselaer Polytechnic Institute.

    All worldwide rights reserved. A license to use, copy, modify and
    distribute this software for noncommercial research purposes only is
    hereby granted, provided that this copyright notice and accompanying
    disclaimer is not modified or removed from the software.

    DISCLAIMER: The software is distributed" AS IS" without any express or
    implied warranty, including but not limited to, any implied warranties
    of merchantability or fitness for a particular purpose or any warrant)'
    of non-infringement of any current or pending patent rights. The authors
    of the software make no representations about the suitability of this
    software for any particular purpose. The entire risk as to the quality
    and performance of the software is with the user. Should the software
    prove defective, the user assumes the cost of all necessary servicing,
    repair or correction. In particular, neither Rensselaer Polytechnic
    Institute, nor the authors of the software are liable for any indirect,
    special, consequential, or incidental damages related to the software,
    to the maximum extent the law permits.
*/

package edu.rpi.cct.uwcal.caldav;

import edu.rpi.cct.webdav.servlet.shared.WebdavIntfException;

/** We map uris onto an object which may be a user, a calendar or an
 * entity contained within that calendar.
 *
 * <p>The response to webdav actions obviously depends upon the type of
 * referenced entity.
 *
 * where user-name needs some predefined value(s) to define public calendar(s)
 * and will be some system defined values to define individual users.
 *
 * <p>The server will have to determine whether a name represents a publicly
 * available calendar or a user and the access to a calendar will, of course,
 * depend upon the authentication state of the user.
 *
 * <p>For development purposes the public calendar name is built in below.
 * For the moment we'll just try to make the public calendar available.
 *
 *   @author Mike Douglass   douglm@rpi.edu
 */
public class CaldavURI {
  public static final int cdURITypeUnknown = 0;
  public static final int cdURITypeUser = 1;
  public static final int cdURITypeCalendar = 2;
  public static final int cdURITypeCalentity = 3;

  /* These need to come from some properties file
   */
  private static String publicCalendarUserName = "public";
  private static String defaultCalendarName = "calendar";

  private String uri;

  String userName;

  String calName;  // Name if it's a calendar
  String id;       // id for an entity

  CaldavURI(String uri, String userName, String calName) {
    reinit(uri, userName, calName);
  }

  /** If id does have ".ics" as a suffix it will be appended to the uri
   */
  CaldavURI(String uri, String userName, String calName, String id) {
    reinit(uri, userName, calName, id);
  }

  void reinit(String uri, String userName, String calName) {
    this.uri = uri;
    this.userName = userName;
    this.calName = calName;
  }

  void reinit(String uri, String userName, String calName, String id) {
    int idLen = 0;
    boolean hasSuffix = false;

    if (id != null) {
      idLen = id.length();
      hasSuffix = id.indexOf(".ics") == (idLen - 4);
    }

    this.userName = userName;
    this.calName = calName;

    if (hasSuffix) {
      this.uri = uri;
      this.id = id.substring(0, idLen - 4);
    } else {
      this.uri = uri + ".ics";
      this.id = id;
    }
  }

  public String getPublicCalendarUserName() {
    return publicCalendarUserName;
  }

  public String getDefaultCalendarName() {
    return defaultCalendarName;
  }

  public boolean getPublic() {
    return publicCalendarUserName.equals(userName);
  }

  public int getType() {
    if (userName == null) {
      return cdURITypeUnknown;
    }

    if (calName == null) {
      return cdURITypeUser;
    }

    if (id == null) {
      return cdURITypeCalendar;
    }

    return cdURITypeCalentity;
  }

  public String getUserName() {
    if (getPublic()) {
      return null;
    }

    return userName;
  }

  public String getCalName() {
    return calName;
  }

  public String getId() {
    return id;
  }

  public String getUri() {
    return uri;
  }

  public boolean isUser() {
    return getType() == cdURITypeUser;
  }

  public boolean isCalendar() {
    return getType() == cdURITypeCalendar;
  }

  public boolean isPublicCalendars() {
    return publicCalendarUserName.equals(userName);
  }

  public boolean isDefaultCalendar() {
    return defaultCalendarName.equals(calName);
  }

  /** Create a new object based on this one with a calendar name appended to the
   * uri. This must be a user uri.
   */
  public CaldavURI appendCalName(String val) throws WebdavIntfException {
    if (getType() != cdURITypeUser) {
      throw new WebdavIntfException("Illegal operation on uri: " + this);
    }

    StringBuffer newUri = new StringBuffer(getUri());
    if (!getUri().endsWith("/")) {
      newUri.append("/");
    }

    newUri.append(val);
    newUri.append("/");

    return new CaldavURI(newUri.toString(), getUserName(), val);
  }

  /** Create a new object based on this one with an event id appended to the
   * uri. This must be a calendar uri.
   */
  public CaldavURI appendId(String val) throws WebdavIntfException {
    if (getType() != cdURITypeCalendar) {
      throw new WebdavIntfException("Illegal operation on uri: " + this);
    }

    StringBuffer newUri = new StringBuffer(getUri());
    if (!getUri().endsWith("/")) {
      newUri.append("/");
    }

    newUri.append(val);

    return new CaldavURI(newUri.toString(), getUserName(), getCalName(), val);
  }

  public String toString() {
    StringBuffer sb = new StringBuffer("CaldavURI{userName=");

    sb.append(String.valueOf(userName));
    sb.append(", calName=");
    sb.append(String.valueOf(calName));
    sb.append(", id=");
    sb.append(String.valueOf(id));
    sb.append("}");

    return sb.toString();
  }
}

