/* **********************************************************************
    Copyright 2005 Rensselaer Polytechnic Institute.

    All worldwide rights reserved. A license to use, copy, modify and
    distribute this software for noncommercial research purposes only is
    hereby granted, provided that this copyright notice and accompanying
    disclaimer is not modified or removed from the software.

    DISCLAIMER: The software is distributed" AS IS" without any express or
    implied warranty, including but not limited to, any implied warranties
    of merchantability or fitness for a particular purpose or any warrant)'
    of non-infringement of any current or pending patent rights. The authors
    of the software make no representations about the suitability of this
    software for any particular purpose. The entire risk as to the quality
    and performance of the software is with the user. Should the software
    prove defective, the user assumes the cost of all necessary servicing,
    repair or correction. In particular, neither Rensselaer Polytechnic
    Institute, nor the authors of the software are liable for any indirect,
    special, consequential, or incidental damages related to the software,
    to the maximum extent the law permits.
*/

package edu.rpi.cct.uwcal.caldav;

import edu.rpi.cct.uwcal.calsvci.CalSvc;
import edu.rpi.cct.uwcal.calsvci.CalSvcI;
import edu.rpi.cct.uwcal.calsvci.CalSvcIPars;
import edu.rpi.cct.uwcal.common.URIgen;
import edu.rpi.cct.uwcal.icalendar.IcalTranslator;

import edu.rpi.cct.uwcal.webcommon.UWCalWebURIgen;

import edu.rpi.cct.webdav.servlet.common.WebdavServlet;
import edu.rpi.cct.webdav.servlet.common.WebdavUtils;
import edu.rpi.cct.webdav.servlet.shared.WebdavIntfException;
import edu.rpi.cct.webdav.servlet.shared.WebdavNsIntf;
import edu.rpi.cct.webdav.servlet.shared.WebdavNsNode;
import edu.rpi.cct.webdav.servlet.shared.WebdavProperty;
import edu.rpi.cct.webdav.servlet.shared.WebdavTags;
import edu.rpi.sss.util.xml.QName;
import edu.rpi.sss.util.xml.XmlEmit;

import edu.washington.cac.calfacade.shared.EventVO;
import edu.washington.cac.calfacade.shared.GuidVO;
import edu.washington.cac.calfacade.shared.UserAuth;

import java.io.Reader;
import java.net.URLEncoder;
import java.util.Collection;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Properties;
import java.util.StringTokenizer;
import java.util.Vector;
import javax.servlet.http.HttpServletRequest;
import org.apache.log4j.Logger;

/** This class implements a namespace interface for the webdav abstract
 * servlet. One of these interfaces is associated with each current session.
 *
 * <p>As a first pass we'll define webdav urls as starting with <br/>
 * /user-name/calendar-name/<br/>
 *
 * <p>uri resolution should be made part of the core calendar allowing all
 * such distinctions to be removed from this code.
 *
 * <p>The part following the above prefix probably determines exactly what is
 * delivered. We may want the entire calendar (or what we show by default) or
 * a single event from the calendar (or does that have the url<br/>
 *   /user-name/event-id.ics<br/>
 *
 *   @author Mike Douglass   douglm@rpi.edu
 */
public class CaldavUWIntf extends WebdavNsIntf {
  /** Namespace prefix based on the request url.
   */
  private String namespacePrefix;

  /** Namespace based on the request url.
   */
  private String namespace;

  /* These two set after a call to getSvci(user)
   */
  private IcalTranslator trans;
  private CalSvcI svci;

  /* for accessing users personal calendar */
  private CalSvcI userSvci;
  private IcalTranslator userTrans;

  /* for accessing public calendars */
  private CalSvcI publicSvci;
  private IcalTranslator publicTrans;

  private URIgen uriGen;

  /* User associated with session */
  private String account;

  /** We store CaldavURI objects here
   */
  private HashMap uriMap = new HashMap();

  /** Set true if we have an authenticated user with sufficient privileges
   */
  private boolean admin;

  /** An object representing the current users access
   */
//  LuwakAccess access;

  /** ===================================================================
   *                     Interface methods
   *  =================================================================== */

  /** Called before any other method is called to allow initialisation to
   * take place at the first or subsequent requests
   */
  public void init(WebdavServlet servlet,
                   HttpServletRequest req,
                   Properties props,
                   XmlEmit xml) throws WebdavIntfException {
    super.init(servlet, req, props, xml);

    namespacePrefix = WebdavUtils.getUrlPrefix(req);
    namespace = namespacePrefix + "/schema";

    uriGen = new UWCalWebURIgen(namespace);

    account = req.getRemoteUser();
  }


  /** Called on the way out to allow resources to be freed.
   */
  public void close() {
    close(userSvci);
    close(publicSvci);
  }

  /** Returns the supported locks for the supportedlocks element.
   * To ensure these will work always provide the full namespace "DAV:"
   * for example, the result for supported exclusive and shared write locks
   * would be the string (formatting not required)
   *
   *  "&lt;DAV:lockentry&gt;" +
   *  "  &lt;DAV:lockscope&gt;" +
   *  "    &lt;DAV:exclusive/&gt;" +
   *  "  &lt;/DAV:lockscope&gt;" +
   *  "  &lt;DAV:locktype&gt;&lt;DAV:write/&gt;&lt;DAV:/locktype&gt;" +
   *  "&lt;/DAV:lockentry&gt;" +
   *  "&lt;DAV:lockentry&gt;" +
   *  "  &lt;DAV:lockscope&gt;" +
   *  "    &lt;DAV:shared/&gt;" +
   *  "  &lt;/DAV:lockscope&gt;" +
   *  "&lt;/DAV:lockentry&gt;"
   */
  public String getSupportedLocks() {
     return  "<DAV:lockentry>" +
             "  <DAV:lockscope>" +
             "    <DAV:exclusive/>" +
             "  </DAV:lockscope>" +
             "  <DAV:locktype><DAV:write/></DAV:locktype>" +
             "</DAV:lockentry>";
  }

  /** Returns true if the namespace supports access control
   */
  public boolean getAccessControl() {
    return true;
  }

  /** Add the namespace for xml tag names in requests and responses.
   * An abbreviation will be supplied by the servlet. The name should be
   * unique in a global sense so don't return something like "DAV:" just
   * because you're David A. Vass
   *
   * <p>Something more like "http://j2ee.rpi.edu:8080/luwak/"
   */
  public void addNamespace() throws WebdavIntfException {
    try {
      xml.addNs(CaldavUWDef.caldavNamespace);
      xml.addNs(CaldavUWDef.icalNamespace);
    } catch (Throwable t) {
      throw new WebdavIntfException(t);
    }
  }

  /** Return the complete URL describing the location of the object
   * represented by the node
   *
   * @param node             node in question
   */
  public String getLocation(WebdavNsNode node) throws WebdavIntfException {
    return namespacePrefix + node.getUri();
  }

  /** Retrieves a node by URI, following any links.
   *
   * @param uri              String URI of the node to retrieve
   * @return WebdavNsNode    node specified by the URI or the node aliased by
   *                         the node at the URI.
   */
  public WebdavNsNode getNode(String uri) throws WebdavIntfException {
    if (uri == null)  {
      return null;
    }

    try {
      CaldavURI wi = findURI(uri);

      if (wi == null) {
        throw new WebdavIntfException("Unable to locate key for " + uri);
      }

      int wit = wi.getType();

      if (wit == CaldavURI.cdURITypeUnknown) {
        throw new WebdavIntfException("Invalid uri: " + uri);
      }

      CaldavUWCalNode nd = null;

      getSvci(wi.getUserName());

      if (wit == CaldavURI.cdURITypeUser) {
        nd = new CaldavUserNode(wi, svci, trans, debug);
      } else if (wit == CaldavURI.cdURITypeCalendar) {
        nd = new CaldavCalNode(wi, svci, trans, debug);
      } else if (wit == CaldavURI.cdURITypeCalentity) {
        nd = new CaldavComponentNode(wi, svci, trans, debug);
      }

      return nd;
    } catch (WebdavIntfException we) {
      throw we;
    } catch (Throwable t) {
      throw new WebdavIntfException(t);
    }
  }

  /** Retrieves a node by URI optionally following the last alias.
   *
   * @param uri              String URI of the node to retrieve
   * @param followAlias      boolean true is we should retrieve the target
   *                         of the alias if the last element is an alias.
   *                         otherwise the node itself is returned
   * @return WebdavNsNode    node specified by the URI or the node aliased by
   *                         the node at the URI if followAlias is true.
   */
  public WebdavNsNode getNode(String uri, boolean followAlias)
      throws WebdavIntfException {
    //XXX probably needs fixing
    return getNode(uri);
  }

  /** Stores/updates an object.
   *
   * @param node             node in question
   */
  public void putNode(WebdavNsNode node)
      throws WebdavIntfException {
  }

  /** Deletes a node from the namespace.
   *
   * @param node             node in question
   */
  public void delete(WebdavNsNode node)
      throws WebdavIntfException {
  }

  /** Returns the immediate children of a node.
   *
   * @param node             node in question
   * @return Iterator        over WebdavNsNode children
   */
  public Iterator getChildren(WebdavNsNode node)
      throws WebdavIntfException {
    try {
      CaldavUWCalNode uwnode = getUwnode(node);

      if (!uwnode.getCollection()) {
        // Don't think we should have been called
        return new Vector().iterator();
      }

      Vector v = new Vector();
      Collection childrenNames = uwnode.getChildren();
      String uri = uwnode.getUri();

      Iterator it = childrenNames.iterator();

      while (it.hasNext()) {
        StringBuffer name = new StringBuffer((String)it.next());
        int pos = 0;

        for (;;) {
          pos = name.indexOf("/", pos);
          if (pos < 0) {
            break;
          }

          name.replace(pos, pos + 1, "-");
        }

        CaldavURI wi = findURI(uri + "/" + name);
        CaldavUWCalNode nd = null;
        getSvci(wi.getUserName());
        int wit = wi.getType();

        if (wit == CaldavURI.cdURITypeCalendar) {
          v.addElement(new CaldavCalNode(wi, svci, trans, debug));
        } else if (wit == CaldavURI.cdURITypeCalentity) {
          v.addElement(new CaldavComponentNode(wi, svci, trans, debug));
        } else {
          throw new WebdavIntfException("Invalid node for getChildren(): " + node);
        }
      }

      return v.iterator();
    } catch (WebdavIntfException we) {
      throw we;
    } catch (Throwable t) {
      throw new WebdavIntfException(t);
    }
  }

  /** Returns the parent of a node.
   *
   * @param node             node in question
   * @return WebdavNsNode    node's parent, or null if the specified node
   *                         is the root
   */
  public WebdavNsNode getParent(WebdavNsNode node)
      throws WebdavIntfException {
    return null;
  }

  /** Entity tags are defined in RFC2068 - they are supposed to provide some
   * sort of indication the data has changed - e.g. a checksum.
   * <p>There are weak and strong tags
   *
   * <p>This methods should return a suitable value for that tag.
   *
   * <p>FIX FIX FIX - not unique?
   */
  public String getEtagValue(WebdavNsNode node, boolean strong)
      throws WebdavIntfException {
    return namespace + String.valueOf(node.getLastmodDate());
  }

  /** Returns all the namespace specific properties for the given node
   * Properties that can be supplied by using node values should not be
   * included.
   *
   * @param node             node in question
   * @return Enumeration     properties
   */
  public Enumeration getProperties(WebdavNsNode node)
      throws WebdavIntfException {
    try {
      CaldavUWCalNode uwnode = getUwnode(node);

      return WebdavProperty.getEnumeration(uwnode.getProperties(namespace));
    } catch (WebdavIntfException we) {
      throw we;
    } catch (Throwable t) {
      throw new WebdavIntfException(t);
    }
  }

  private void addProp(Vector v, QName tag, Object val) {
    if (val != null) {
      v.addElement(new WebdavProperty(tag, String.valueOf(val)));
    }
  }

  /** Returns an InputStream for the content.
   *
   * @param node             node in question
   * @return Reader          A reader for the content or null.
   */
  public Reader getContent(WebdavNsNode node)
      throws WebdavIntfException {
    try {
      if (!node.getAllowsGet()) {
        return null;
      }

      CaldavUWCalNode uwnode = getUwnode(node);

      return uwnode.getContent();
    } catch (WebdavIntfException we) {
      throw we;
    } catch (Throwable t) {
      throw new WebdavIntfException(t);
    }
  }

  /** Set the content from a Reader
   *
   * @param node             node in question.
   * @param content          String body value.
   * @param create           true if this is a creation
   * @return WebdavNsNode    possibly modified node referring to new object
   */
  public WebdavNsNode putContent(WebdavNsNode node, String content,
                                 boolean create)
      throws WebdavIntfException {
    try {
      CaldavUWCalNode uwnode = getUwnode(node);
      CaldavURI cdUri = uwnode.getCDURI();
      String idStr = cdUri.getId();
      boolean addName = false;

      Collection c = trans.toEvent(content);
      if (c.size() != 1) {
        throw new WebdavIntfException("Bad request");
      }

      EventVO ev = (EventVO)c.iterator().next();
      EventVO oldev = null;

      // See if the id is one of ours
      GuidVO guid = GuidVO.valueOf(idStr);
      int evid = 0;

      if (svci.isInternalGuid(guid)) {
        evid = guid.getId();
        if (debug) {
          debugMsg("Look up event with id " + evid);
        }
      } else {
        // Look up the id
        evid = svci.lookupEventName(idStr);
        addName = true;  // for create
        if (debug) {
          debugMsg("Look up event with name " + idStr + " and get " + evid);
        }
      }

      if (evid <= 0) {
        create = true;
      } else {
        /* See if the event exists */
        oldev = svci.getEvent(evid);
        create = oldev == null;
      }

      if (!create) {
        svci.updateEvent(ev, oldev);
      } else {
        svci.addEvent(ev);

        if (addName) {
          svci.setEventName(idStr, ev.getId());
        }

        // Create a new node to represent this event
        node = getNode("/" + cdUri.getUserName() +
                       "/" + cdUri.getCalName() +
                       "/" + idStr + ".ics");
      }

      return node;
    } catch (WebdavIntfException we) {
      throw we;
    } catch (Throwable t) {
      throw new WebdavIntfException(t);
    }
  }

  /** Create a new node.
   *
   * @param node             node to create with new uri set
   */
  public void create(WebdavNsNode node)
      throws WebdavIntfException {
  }

  /** Creates an alias to another node.
   *
   * @param alias       alias node that should be created with uri and
   *                    targetUri set
   */
  public void createAlias(WebdavNsNode alias)
      throws WebdavIntfException {
  }

  /* ====================================================================
   *                Property value methods
   * ==================================================================== */

  /** Generate default webdav resource type.
   *
   * @param node      for which we are generating resource type
   */
  public void generatePropResourcetype(WebdavNsNode node)
          throws WebdavIntfException {
    CaldavUWCalNode uwnode = getUwnode(node);

    if (uwnode.isCalendar()) {
      try {
        xml.openTag(WebdavTags.resourcetype);
        xml.emptyTag(WebdavTags.collection);
        xml.emptyTag(CaldavTags.calendar);
        xml.closeTag(WebdavTags.resourcetype);
      } catch (Throwable t) {
        throw new WebdavIntfException(t);
      }
    } else {
      super.generatePropResourcetype(node);
    }
  }

  /* ====================================================================
   *                         Private methods
   * ==================================================================== */

  private CalSvcI getSvci(String user) throws WebdavIntfException {
    boolean publicMode = user == null;

    if (publicMode) {
      svci = publicSvci;
    } else {
      svci = userSvci;
    }

    if (svci != null) {
      if (!svci.isOpen()) {
        try {
          svci.open();
          svci.beginTransaction();
        } catch (Throwable t) {
          throw new WebdavIntfException(t);
        }
      }

      return svci;
    }

    try {
      svci = new CalSvc();
      CalSvcIPars pars = new CalSvcIPars(account, UserAuth.noPrivileges,
                                         account,
                                         publicMode,
                                         null, // synchId
                                         false, false, false, false,
                                         debug);
      svci.init(pars);

      svci.open();
      svci.beginTransaction();

      trans = new IcalTranslator(svci, uriGen);
    } catch (Throwable t) {
      throw new WebdavIntfException(t);
    }

    if (publicMode) {
      publicSvci = svci;
      publicTrans = trans;
    } else {
      userSvci = svci;
      userTrans = trans;
    }

    return svci;
  }

  private void close(CalSvcI svci) {
    if ((svci == null) || !svci.isOpen()) {
      return;
    }

    try {
      svci.endTransaction();
    } catch (Throwable t) {
    }

    try {
      svci.close();
    } catch (Throwable t) {
    }
  }

  /*
  private void setAccess(HttpServletRequest request,
                         LuwakState ks) {
    access = ks.getAccess();

    if (access.getAccessSet()) {
      return;
    }

    access.init(ks.getUser(),
            request.isUserInRole(
              props.getProperty("edu.rpi.cct.luwak.role.admin")),
            request.isUserInRole(
              props.getProperty("edu.rpi.cct.luwak.role.chiefeditor")),
            request.isUserInRole(
              props.getProperty("edu.rpi.cct.luwak.role.editor")),
            request.isUserInRole(
              props.getProperty("edu.rpi.cct.luwak.role.writer")),
            request.isUserInRole(
              props.getProperty("edu.rpi.cct.luwak.role.broadcaster")),
            request.isUserInRole(
              props.getProperty("edu.rpi.cct.luwak.role.eventpublisher")),
            debug);
    /** security-role-ref in web.xml doesn't seem to work (or I missed something
        somewhere). Until it does work use the above
    access.setAdminUser(request.isUserInRole("admin"));
    access.setWriterUser(request.isUserInRole("writer")); * /
  }*/

  /** Find the named item by following down the path from the root.
   * This requires the names at each level to be unique (and present)
   *
   * <p>NOTE: from caldav-04   <br/>
   * For optimum interoperability with existing HTTP clients, CalDAV
   * clients and servers MUST use the file extension ".ics" as well as the
   " text/calendar" MIME type, whenever creating Calendar objects of that
   * MIME type.
   *
   * <p>For the moment assume /user-name/calendar-name/event-guid.ics or <br/>
   * /user-name/calendar-name or <br/>
   * /user-name  <br/>
   *
   */
  private CaldavURI findURI(String uri) throws WebdavIntfException {
    try {
      String[] ss = uri.split("/");
      StringBuffer curUri = new StringBuffer("/");
      int pathLength = 0;
      String[] pathElements = new String[3];
      String userName = null;
      String calendarName = null;

      mapUriName(curUri.toString(), null, null);

      for (int i = 0; i < ss.length; i++) {
        String s = ss[i];

        if ((s == null) || (s.length() == 0) || (s.equals("."))) {
          // ignore
        } else if (s.equals("..")) {
          // Back up 1
          if (pathLength == 0) {
            // back too far
            return null;
          }

          pathLength--;
        } else if (pathLength == 0) {
          /* Expect a user-name
           */

          userName = ss[i];
          curUri.append(userName);
          mapUriName(curUri.toString(), userName, null);
          pathLength++;
        } else if (pathLength == 1) {
          /* Expect a calendar name
           */

          calendarName = ss[i];
          curUri.append("/");
          curUri.append(calendarName);
          mapUriName(curUri.toString(), userName, calendarName);
          pathLength++;
        } else if (pathLength == 2) {
          /* Expect an event id or something????
           */
          curUri.append("/");
          curUri.append(ss[i]);
          mapUriId(curUri.toString(), userName, calendarName, ss[i]);
          pathLength++;
        } else {
          throw new WebdavIntfException("Invalid URI");
        }
      }

      return (CaldavURI)uriMap.get(curUri.toString());
    } catch (Throwable t) {
      throw new WebdavIntfException(t);
    }
  }

  private void mapUriName(String uri, String userName, String calName) {
    CaldavURI wi = (CaldavURI)uriMap.get(uri);

    if (wi != null) {
      wi.reinit(uri, userName, calName);
      return;
    }

    uriMap.put(uri, new CaldavURI(uri, userName, calName));
  }

  private void mapUriId(String uri, String userName, String calName, String id) {
    CaldavURI wi = (CaldavURI)uriMap.get(uri);

    if (wi != null) {
      wi.reinit(uri, userName, calName, id);
      return;
    }

    uriMap.put(uri, new CaldavURI(uri, userName, calName, id));
  }

  private CaldavUWCalNode getUwnode(WebdavNsNode node)
      throws WebdavIntfException {
    if (!(node instanceof CaldavUWCalNode)) {
      throw new WebdavIntfException("Not a valid node object " +
                                    node.getClass().getName());
    }

    return (CaldavUWCalNode)node;
  }
}

