/* **********************************************************************
    Copyright 2005 Rensselaer Polytechnic Institute.

    All worldwide rights reserved. A license to use, copy, modify and
    distribute this software for noncommercial research purposes only is
    hereby granted, provided that this copyright notice and accompanying
    disclaimer is not modified or removed from the software.

    DISCLAIMER: The software is distributed" AS IS" without any express or
    implied warranty, including but not limited to, any implied warranties
    of merchantability or fitness for a particular purpose or any warrant)'
    of non-infringement of any current or pending patent rights. The authors
    of the software make no representations about the suitability of this
    software for any particular purpose. The entire risk as to the quality
    and performance of the software is with the user. Should the software
    prove defective, the user assumes the cost of all necessary servicing,
    repair or correction. In particular, neither Rensselaer Polytechnic
    Institute, nor the authors of the software are liable for any indirect,
    special, consequential, or incidental damages related to the software,
    to the maximum extent the law permits.
*/

package edu.rpi.cct.uwcal.caldav;

import edu.rpi.cct.webdav.servlet.common.MethodBase;
import edu.rpi.cct.webdav.servlet.shared.WebdavException;
import edu.rpi.sss.util.xml.XmlUtil;

import java.text.ParseException;
import java.util.Date;
import java.util.Iterator;
import java.util.Vector;
import javax.servlet.http.HttpServletResponse;

import net.fortuna.ical4j.util.DateFormat;

import org.apache.log4j.Logger;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;

/** Class to represent a calendar-query calendar-data element
 *
 *   @author Mike Douglass   douglm@rpi.edu
 */
public class CalendarQueryCalendarData {
  /*
      <!ELEMENT calendar-data ((comp expand-recurrence-set?) |
                    #PCDATA)?>

         pcdata is for response

      <!ATTLIST calendar-data return-content-type CDATA
                      "text/calendar">

      <!ELEMENT comp ((allcomp, (allprop | prop*)) |
                       (comp*, (allprop | prop*)))>

      <!ATTLIST comp name CDATA #REQUIRED>

      <!ELEMENT allcomp EMPTY>

      <!ELEMENT allprop EMPTY>

      <!ELEMENT prop EMPTY>

      <!ATTLIST prop name CDATA #REQUIRED
                     novalue (yes|no) "no">

      <!ELEMENT expand-recurrence-set EMPTY>

      <!ATTLIST expand-recurrence-set start CDATA #REQUIRED
                                      end CDATA #REQUIRED>

   */
  private boolean debug;

  protected transient Logger log;

  private String returnContentType; // null for defaulted
  private Comp comp;
  private ExpandRecurrenceSet ers;

  /* Set by internal methods to indicate failure reason.
   */
  private int status;

  public CalendarQueryCalendarData(boolean debug) {
    this.debug = debug;
  }

  /** The given node must be the Filter element
   *
   * @return int     htp status
   */
  public int parse(Node nd) throws WebdavException {
    /* Either empty - show everything or
              comp + optional expand-recurrence-set
     */
    NamedNodeMap nnm = nd.getAttributes();

    if ((nnm != null) && (nnm.getLength() > 1)) {
      return HttpServletResponse.SC_BAD_REQUEST;
    }

    if (nnm != null) {
      if (nnm.getLength() == 1) {
        returnContentType = XmlUtil.getAttrVal(nnm, "return-content-type");
        if (returnContentType == null) {
          return HttpServletResponse.SC_BAD_REQUEST;
        }
      } else if (nnm.getLength() > 0) {
        // Bad attribute(s)
        return HttpServletResponse.SC_BAD_REQUEST;
      }
    }

    Element[] children = getChildren(nd);

    for (int i = 0; i < children.length; i++) {
      Node curnode = children[i];

      if (debug) {
        trace("calendar-data node type: " +
              curnode.getNodeType() + " name:" +
              curnode.getNodeName());
      }

      if (MethodBase.nodeMatches(curnode, CaldavTags.comp)) {
        if (comp != null) {
          return HttpServletResponse.SC_BAD_REQUEST;
        }

        comp = parseComp(curnode);

        if (comp == null) {
          return status;
        }
      } else if (MethodBase.nodeMatches(curnode, CaldavTags.expandRecurrenceSet)) {
        if (ers != null) {
          return HttpServletResponse.SC_BAD_REQUEST;
        }

        ers = parseExpandRecurrenceSet(curnode);

        if (ers == null) {
          return status;
        }
      } else {
        return HttpServletResponse.SC_BAD_REQUEST;
      }
    }

    return HttpServletResponse.SC_OK;
  }

  private Comp parseComp(Node nd) throws WebdavException {
    /* Either allcomp + (either allprop or 0 or more prop) or
              0 or more comp + (either allprop or 0 or more prop)
     */
    Comp c = new Comp();

    c.name = getOnlyAttribute(nd, "name");
    if (c.name == null) {
      status = HttpServletResponse.SC_BAD_REQUEST;
      return null;
    }

    Element[] children = getChildren(nd);

    for (int i = 0; i < children.length; i++) {
      Node curnode = children[i];

      if (debug) {
        trace("comp node type: " +
              curnode.getNodeType() + " name:" +
              curnode.getNodeName());
      }

      if (MethodBase.nodeMatches(curnode, CaldavTags.allcomp)) {
        if (c.comps != null) {
          status = HttpServletResponse.SC_BAD_REQUEST;
          return null;
        }

        c.allcomp = true;
      } else if (MethodBase.nodeMatches(curnode, CaldavTags.comp)) {
        if (c.allcomp) {
          status = HttpServletResponse.SC_BAD_REQUEST;
          return null;
        }

        Comp chc = parseComp(curnode);

        if (chc == null) {
          return null;
        }

        if (c.comps == null) {
          c.comps = new Vector();
        }

        c.comps.add(chc);
      } else if (MethodBase.nodeMatches(curnode, CaldavTags.allprop)) {
        if (c.props != null) {
          status = HttpServletResponse.SC_BAD_REQUEST;
          return null;
        }

        c.allprop = true;
      } else if (MethodBase.nodeMatches(curnode, CaldavTags.prop)) {
        if (c.allprop) {
          status = HttpServletResponse.SC_BAD_REQUEST;
          return null;
        }

        Prop pr = parseProp(curnode);

        if (pr == null) {
          return null;
        }

        if (c.props == null) {
          c.props = new Vector();
        }

        c.props.add(pr);
      } else {
        status = HttpServletResponse.SC_BAD_REQUEST;
        return null;
      }
    }

    return c;
  }

  private ExpandRecurrenceSet parseExpandRecurrenceSet(Node nd)
          throws WebdavException {
    NamedNodeMap nnm = nd.getAttributes();

    ExpandRecurrenceSet ers = new ExpandRecurrenceSet();

    if ((nnm == null) || (nnm.getLength() != 2)) {
      status = HttpServletResponse.SC_BAD_REQUEST;
      return null;
    }

    try {
      DateFormat df = DateFormat.getInstance();
      Node nmAttr = nnm.getNamedItem("start");

      if (nmAttr == null) {
        status = HttpServletResponse.SC_BAD_REQUEST;
        return null;
      }

      ers.start = df.parse(nmAttr.getNodeValue());

      nmAttr = nnm.getNamedItem("end");

      if (nmAttr == null) {
        status = HttpServletResponse.SC_BAD_REQUEST;
        return null;
      }

      ers.end = df.parse(nmAttr.getNodeValue());
    } catch (ParseException pe) {
      status = HttpServletResponse.SC_BAD_REQUEST;
      return null;
    }

    return ers;
  }

  private Prop parseProp(Node nd) throws WebdavException {
    NamedNodeMap nnm = nd.getAttributes();

    Prop pr = new Prop();

    if ((nnm == null) || (nnm.getLength() == 0)) {
      status = HttpServletResponse.SC_BAD_REQUEST;
      return null;
    }

    int attrCt = nnm.getLength();

    pr.name = XmlUtil.getAttrVal(nnm, "name");
    if (pr.name == null) {
      status = HttpServletResponse.SC_BAD_REQUEST;
      return null;
    }

    attrCt--;

    Boolean val = null;

    try {
      val = XmlUtil.getYesNoAttrVal(nnm, "novalue");
    } catch (Throwable t) {
      status = HttpServletResponse.SC_BAD_REQUEST;
      return null;
    }

    if (val != null) {
      pr.novalue = val.booleanValue();
    }

    return pr;
  }

  private Element[] getChildren(Node nd) {
    try {
      return XmlUtil.getElementsArray(nd);
    } catch (Throwable t) {
      if (debug) {
        getLogger().error("<filter>: parse exception: ", t);
      }

      status = HttpServletResponse.SC_BAD_REQUEST;
      return null;
    }
  }

  private Element getOnlyChild(Node nd) {
    try {
      return XmlUtil.getOnlyElement(nd);
    } catch (Throwable t) {
      if (debug) {
        getLogger().error("<filter>: parse exception: ", t);
      }

      status = HttpServletResponse.SC_BAD_REQUEST;
      return null;
    }
  }

  private String getOnlyAttribute(Node nd, String name) {
    NamedNodeMap nnm = nd.getAttributes();

    if ((nnm == null) || (nnm.getLength() != 1)) {
      status = HttpServletResponse.SC_BAD_REQUEST;
      return null;
    }

    String val = XmlUtil.getAttrVal(nnm, "name");
    if (val == null) {
      status = HttpServletResponse.SC_BAD_REQUEST;
      return null;
    }

    return val;
  }

  /** ===================================================================
   *                   Dump methods
   *  =================================================================== */

  public void dump() {
    StringBuffer sb = new StringBuffer("  <calendar-data");

    if (returnContentType != null) {
      sb.append("  return-content-type=\"");
      sb.append(returnContentType);
      sb.append("\"");
    }
    sb.append(">");
    trace(sb.toString());

    if (comp != null) {
      comp.dump(getLogger(), "    ");
    }

    if (ers != null) {
      ers.dump(getLogger(), "    ");
    }

    trace("  </calendar-data>");
  }

  /** ===================================================================
   *                   Logging methods
   *  =================================================================== */

  protected Logger getLogger() {
    if (log == null) {
      log = Logger.getLogger(this.getClass());
    }

    return log;
  }

  protected void debugMsg(String msg) {
    getLogger().debug(msg);
  }

  protected void logIt(String msg) {
    getLogger().info(msg);
  }

  protected void trace(String msg) {
    getLogger().debug(msg);
  }

  /** ===================================================================
   *                   Private classes
   *  =================================================================== */

  private static class Comp {
    String name;

    /* if allcomp != true then look at comp */
    boolean allcomp;

    Vector comps; // null for zero otherwise vector of comp

    // true or look at props
    boolean allprop;

    Vector props; // null for zero otherwise vector of prop

    public void dump(Logger log, String indent) {
      StringBuffer sb = new StringBuffer(indent);

      sb.append("<comp name=");
      sb.append(name);
      sb.append(">");
      log.debug(sb.toString());

      if (allcomp) {
        log.debug(indent + "  <allcomp/>");
      } else if (comps != null) {
        Iterator it = comps.iterator();

        while (it.hasNext()) {
          Comp c = (Comp)it.next();
          c.dump(log, indent + "  ");
        }
      }

      if (allprop) {
        log.debug(indent + "  <allprop/>");
      } else if (props != null) {
        Iterator it = props.iterator();

        while (it.hasNext()) {
          Prop pr = (Prop)it.next();
          pr.dump(log, indent + "  ");
        }
      }

      log.debug(indent + "</comp>");
    }
  }

  private static class ExpandRecurrenceSet {
    Date start;
    Date end;

    public void dump(Logger log, String indent) {
      StringBuffer sb = new StringBuffer(indent);

      sb.append("<expand-recurrence-set start=");
      sb.append(start);
      sb.append(" end=");
      sb.append(end);
      sb.append("/>");

      log.debug(sb.toString());
    }
  }

  private static class Prop {
    String name;
    boolean novalue;

    public void dump(Logger log, String indent) {
      StringBuffer sb = new StringBuffer(indent);

      sb.append("<calddav:prop name=");
      sb.append(name);
      sb.append(" novalue=");
      sb.append(novalue);
      sb.append("/>");

      log.debug(sb.toString());
    }
  }
}

